/*
    decode.cpp
    The Latent Words Language Model, decoder.
    $Id: decode.cpp,v 1.5 2010/03/18 08:17:29 daichi Exp $

*/
#include "lwlm.h"
#include "util.h"
#include "random.h"
#include "debug.h"
#include <vector>
#include <cstdio>
#include <algorithm>

using namespace std;

struct cell {
	int word;
	int count;
};

int
cellsort (cell *a, cell *b)
{
	return a->count > b->count;
}

void
LWLM::write_decoded (vector<vector<int> *> &sampled, vector<int> &s,
		     FILE *fp, char *fmt, double epsilon, int K)
{
	static vector<cell *> appears;
	int i, c, n, t, v, w, size = sampled.size();
	double p;

	if (appears.empty())
	{
		while (appears.size() < L)
		{
			appears.push_back (new cell);
		}
	}

	for (t = 0; t < size; t++)
	{
		for (v = 0; v < L; v++)
		{
			appears[v]->word = v;
			appears[v]->count = 0;
		}
		
		if ((n = sampled[t]->size()) > 0)
		{
			fprintf(fp, "%s\t", ids(s[t]));
			for (i = 0; i < n; i++)
			{
				w = (*sampled[t])[i];
				appears[w]->count++;
			}
			sort (appears.begin(), appears.end(), cellsort);
			for (v = 0; v < L; v++)
			{
				w = appears[v]->word;
				c = appears[v]->count;
				p = c / (double) n;
				
				if (c == 0) break;
				if (v >= K) break;
				if (p < epsilon) break;
				
				fprintf(fp, fmt, ids(w), p);
			}
			fprintf(fp, "\n");
		}
	}
}

void
LWLM::gibbs_decode (vector<vector<int> *> &sampled, vector<int> &s,
		    int burn, int interval)
{
	static vector<int> width, hidden, orders;
	int i, j, t, size = s.size();

	/* buffers */
	while (sampled.size() < size)
	{
		sampled.push_back (new vector<int>);
	}
	width.resize (size);
	orders.resize (size);
	sampled.resize (size);
	for (i = 0; i < size; i++)
		sampled[i]->resize (0);

	/* prepare */
	hidden = s;   /* copy */
	for (t = order; t < size - 1; t++)
	{
		if (t + order < size)
			width[t] = order + 1;
		else
			width[t] = size - t;
	}
	for (t = 0; t < size; t++)
		orders[t] = t;

	/* burn-in */
	for (j = 0; j < burn; j++)
	{
		eprintf("burn-in %2d/%2d..\r", j + 1, burn);
		random_shuffle (orders.begin(), orders.end(), irand);
		for (i = 0; i < size; i++)
		{
			t = orders[i];
			if (s[t] != EOS)
			{
				hidden[t] = slice_gibbs_draw (&hidden[t], &s[t], width[t]);
				// hidden[t] = gibbs_draw (&hidden[t], &s[t], width[t]);
			}
		}
	}
	/* sample */
	for (j = 0; j < iter; j++)
	{
		eprintf("sampling %2d/%2d..\r", j + 1, iter);
		random_shuffle (orders.begin(), orders.end(), irand);
		for (i = 0; i < size; i++)
		{
			t = orders[i];
			if (s[t] != EOS)
			{
				hidden[t] = slice_gibbs_draw (&hidden[t], &s[t], width[t]);
				// hidden[t] = gibbs_draw (&hidden[t], &s[t], width[t]);
				if (j % interval == 0)
				{
					sampled[t]->push_back (hidden[t]);
				}
			}
		}
	}
	eprintf("sampling %2d/%2d..done.\n", j, iter);
}
				
void
LWLM::decode (char *out, char *fmt, int burn, int interval,
	      double epsilon, int K)
{
	static vector<vector<int> *> sampled;
	int size, i, n = observed.size();
	vector<int> *s;
	witerator w;
	FILE *fp;

	if (out == NULL)
		fp = stdout;
	else {
		if ((fp = fopen(out, "w")) == NULL)
		{
			perror("fopen");
			return;
		}
	}

	for (i = 0; i < n; i++)
	{
		s = observed[i];
		size = s->size();
		eprintf("decoding sentence %3d/%d..\n", i + 1, n);

		gibbs_decode (sampled, *s, burn, interval);
		write_decoded (sampled, *s, fp, fmt, epsilon, K);
		
		fprintf(fp, "\n");  /* separate sentence */
	}
	eprintf("decoding sentence %3d/%d..done.\n", i, n);

	if (fp != stdout) fclose(fp);

}

