/*
    hpylm.h
    $Id: hpylm.h,v 1.5 2010/03/17 16:41:57 daichi Exp $
    
 */
#ifndef _HPYLM_H_
#define _HPYLM_H_
#include <vector>
#include <utility>
#include "hash.h"
#include "record.h"
#define  EOS    -1
#define  cons	make_pair

using namespace std;

typedef struct HPYLM HPYLM;
typedef struct ngram ngram;
typedef struct restaurant restaurant;
typedef vector<int>::iterator witerator;

/*
 *  HPYLM definitions
 *
 */

struct restaurant {	/* c_uwx */
	int word;
	int cur;	/* current position in record */
	int ncounts;
	ngram *context;
	vector<int> tables;

	/* interface */
	bool increment_customer ();
	void increase_table ();
	void increment_table (int k);

	bool decrement_customer ();
	void decrease_table (int k);
	void decrement_table (int k);

	/* const/destructors */
	restaurant (int word_, ngram *context_);
	restaurant (ngram *context_);
	~restaurant ();
};

struct ngram {		/* c_uxx */
	int word;
	hashmap *words;
	hashmap *children;
	record *records;
	ngram *parent;
	
	int level;
	int ncounts;
	int ntables;

	/* common */
	static HPYLM *lm;
	
	/* interface */
	bool increment_customer (int w);
	bool decrement_customer (int w);

	/* probability */
	double hpy_probability (int w);

	/* const/destructors */
	ngram (ngram *parent_, int word_, HPYLM *lm_);
	ngram (ngram *parent_, int word_);
	ngram ();
	~ngram ();

	/* print */
	void print ();
	void print_node ();

	/* misc */
	double hpy_residuals ();
	
};

/*
 *  HPYLM language model class
 *
 */

struct HPYLM {
	int order;
	int lexicon;
	ngram *tree;
	vector<double> d;
	vector<double> theta;
	vector<witerator> words;

	/* interfaces */
	bool add_data (witerator w);
	bool add_customer (witerator w);
	bool remove_customer (witerator w);
	bool _add_customer (int n, witerator word, witerator history, ngram *context);
	bool _remove_customer (int n, witerator word, witerator history, ngram *context);
	
	/* inference */
	bool gibbs (int iter);
	bool hyper_estimate ();
	double hpy_d_posterior (int n);
	double hpy_theta_posterior (int n);
	double hpy_d_expectation (int n);
	double hpy_theta_expectation (int n);

	/* probability */
	double hpy_probability (int *word, int *history);
	double hpy_probability (witerator word, witerator history);
	double _hpy_probability (int n, witerator word, witerator history, ngram *context);
	
	/* misc */
	int customers ();
	ngram *find_context (witerator history);
	ngram *_find_context (int n, witerator history, ngram *context);

	/* I/O */
	void save (char *file);
	void load (char *file);
	
	/* print */
	void print ();
	
	HPYLM (int order_);
	~HPYLM ();
};




#endif

