/*
    hyper.cpp
    $Id: hyper.cpp,v 1.2 2010/03/08 11:22:22 daichi Exp $

*/
#include <cstdio>
#include <cmath>
#include "hyper.h"
#include "hpylm.h"
#include "hash.h"
#include "random.h"

static double s0[] = { 1, 1 };	/* gamma prior */
static double b0[] = { 1, 1 };	/* beta prior */

/* internal functions */

typedef struct nparam nparam;
struct nparam {
	int n;
	HPYLM *lm;
};

static double hpy_yi (ngram *node, nparam *arg);
static double hpy_yi_inv (ngram *node, nparam *arg);
static double hpy_logx (ngram *node, nparam *arg);
static double hpy_zwkj_inv (ngram *node, nparam *arg);
static double word_zwkj_inv (restaurant *word, double *dp);

/* implementations */

double
HPYLM::hpy_theta_posterior (int n)
{
	nparam arg;
	arg.n = n; arg.lm = this;
	return gamrand (s0[0] + hpy_yi (tree, &arg),
			s0[1] - hpy_logx (tree, &arg));
}

double
HPYLM::hpy_d_posterior (int n)
{
	nparam arg;
	arg.n = n; arg.lm = this;
	return betarand (b0[0] + hpy_yi_inv (tree, &arg),
			 b0[1] + hpy_zwkj_inv (tree, &arg));
}

double
HPYLM::hpy_theta_expectation (int n)
{
	nparam arg;
	arg.n = n; arg.lm = this;
	return (s0[0] + hpy_yi (tree, &arg))
	     / (s0[1] - hpy_logx (tree, &arg));
}

double
HPYLM::hpy_d_expectation (int n)
{
	double a, b;
	nparam arg;
	arg.n = n; arg.lm = this;
	a = b0[0] + hpy_yi_inv (tree, &arg);
	b = b0[1] + hpy_zwkj_inv (tree, &arg);
	return a / (a + b);
}

static double
hpy_yi (ngram *node, nparam *arg)
{
	int i, n = arg->n;
	HPYLM *lm = arg->lm;
	double d, theta, z = 0;

	if (node->level < n)
	{
		return hash_fold (node->children, (foldfunc) hpy_yi, arg);
	}
	else {	/* node->level == n */
		if (node->ntables > 1)
		{
			d = lm->d[n];
			theta = lm->theta[n];
			for (i = 1; i < node->ntables; i++)
				z += bernoulli (theta / (theta + d * i));
		}
		return z;
	}
}

static double
hpy_yi_inv (ngram *node, nparam *arg)
{
	int i, n = arg->n;
	HPYLM *lm = arg->lm;
	double d, theta, z = 0;
	
	if (node->level < n)
	{
		return hash_fold (node->children, (foldfunc) hpy_yi_inv, arg);
	}
	else {
		if (node->ntables > 1) {
			d = lm->d[n];
			theta = lm->theta[n];
			for (i = 1; i < node->ntables; i++)
				z += 1 - bernoulli(theta / (theta + d * i));
		}
		return z;
	}
}

static double
hpy_logx (ngram *node, nparam *arg)
{
	int n = arg->n;
	HPYLM *lm = arg->lm;
	double theta;
	
	if (node->level < n)
	{
		return hash_fold (node->children, (foldfunc) hpy_logx, arg);
	}
	else {
		if (node->ncounts > 1) {
			theta = lm->theta[n];
			return log(betarand(theta + 1, (double)node->ncounts - 1));
		} else
			return 0;
	}
}

static double
hpy_zwkj_inv (ngram *node, nparam *arg)
{
	int n = arg->n;
	HPYLM *lm = arg->lm;
	
	if (node->level < n)
		return hash_fold (node->children, (foldfunc)hpy_zwkj_inv, arg);
	else
		return hash_fold (node->words, (foldfunc)word_zwkj_inv, &(lm->d[n]));
}

static double
word_zwkj_inv (restaurant *word, double *dp)
{
	int j, k, cuwk;
	double z = 0;

	for (k = 0; k < word->tables.size(); k++)
	{
		cuwk = word->tables[k];
		if (cuwk > 1)
		{
			for (j = 1; j < cuwk; j++)
				z += 1 - bernoulli ((j - 1) / (j - *dp));
		}
	}
	return z;
}

